import java.sql.*;
import java.util.Iterator;

public class VierGewinntDB{
    private String connectionString;
    private int generationCount = 0;
   
    //JDBC stuff
    private Connection con;
    private Statement statement;

    //how much entries should be send with a single insert statement
    private final static int MAXQLOAD = VP.MAXDBSTORE;
    private VierGewinntTurn[] insertQ = new VierGewinntTurn[MAXQLOAD];
    private int insertQLoad = 0;

    //the upper limit of the number of entries to select in a single query  
    private final static int MAXFETCH = VP.MAXDBFETCH;
    
    //Does the execution of SQL statements in the background
    private final SQLExecuter sqlExecuter = new SQLExecuter();

    //stores the time used so far, for load balancing
    private long timeUsed = 0;

    //the sql statement for inserting, cashes the memory
    private final StringBuffer insertSQL = new StringBuffer(MAXQLOAD*2*20);

    private final static boolean MULTITHREADED = VP.MULTITHREADED; 

    private class SQLExecuter extends Thread{
	private String sql;
	private boolean isRunning = false;

	public SQLExecuter(){
	    super();
	    setPriority(Thread.MAX_PRIORITY);
	    setDaemon(true); //application terminates if run hasn't terminated
	    start();
	}
	
	public synchronized void executeSQL(String sql){
	    this.sql= sql;
	    isRunning = true;
	    notifyAll();
	}

	public synchronized void run(){
	    while(true){
		try{
		    wait();
		}catch(InterruptedException e){
		    e.printStackTrace();
		}
		
		long startTime = System.currentTimeMillis();

		try{
		    statement.executeUpdate(sql);
		}catch(SQLException e){
		    System.err.println("ConnectionString='"
				       +connectionString+"'");
		    e.printStackTrace();
		    System.exit(-1);
		}

		isRunning=false;

		long endTime = System.currentTimeMillis();
		timeUsed += endTime-startTime;
		notifyAll();
	    }
	}
	
	public synchronized void waitToFinish(){
	    if(isRunning){
		System.out.println("waiting["+connectionString.substring(0,10)
				   +"](");
		try{
		    wait();
		}catch(InterruptedException e){
		    e.printStackTrace();
		}
		System.out.println(")");
	    }
	}
    }

    public VierGewinntDB (String connectionString){
    	this.connectionString = connectionString;
	try {
	    con = DriverManager.getConnection(connectionString);
	    statement = con.createStatement();
	}catch(SQLException e){
	    System.err.println("ConnectionString='"+connectionString+"'");
	    e.printStackTrace();
	    System.exit(-1);
	}
	

	dropAllTables();
	createTable();
    }
    
    public String toString(){
	return super.toString()+":"+connectionString;
    }

    public long getTimeUsedAndReset(){
	long ret = timeUsed;
	timeUsed = 0;
	return ret;
    }

    private void dropAllTables(){
	for(int i=0;i<VP.MAXTURNDEPTH;i++)
	    dropTable("Generation"+i);
    }

    private void dropTable(String tableName){
	try{
	    statement.executeUpdate("DROP TABLE "+tableName);
	}catch(SQLException e){
	    //do nothing, nothing more to delete
	}
    }

    private void createTable(){
	final String tableName = "Generation"+generationCount;
	dropTable(tableName);
	String sql = "CREATE TABLE "+tableName+" ("
	    //	    +"ID Integer NOT NULL AUTO_INCREMENT PRIMARY KEY,"
	    +"data1 BigInt NOT NULL,"
	    +"data2 BigInt NOT NULL,"
	    +"PRIMARY KEY (data1,data2)"
	    +")";
	
	long startTime = System.currentTimeMillis();
	try{
	    statement.executeUpdate(sql);	
	}catch(SQLException e){
	    System.err.println("ConnectionString='"+connectionString+"'");
	    e.printStackTrace();
	    System.exit(-1);
	}
	long endTime = System.currentTimeMillis();
	timeUsed += endTime-startTime;
    }

    public void open(){
	long startTime = System.currentTimeMillis();
	try{
	    String sql = "LOCK TABLE "
		+"Generation"+generationCount+" WRITE";
	    if(generationCount>0)
		sql += ", Generation"+(generationCount-1)+" READ"; 

	    statement.executeUpdate(sql);
	}catch(SQLException e){
	    System.err.println("ConnectionString='"+connectionString+"'");
	    e.printStackTrace();
	    System.exit(-1);
	}
	long endTime = System.currentTimeMillis();
	timeUsed += endTime-startTime;
    }

    public void close(){
	flushData();
	sqlExecuter.waitToFinish();

	long startTime = System.currentTimeMillis();
	try{
	    if(!VP.STORETREE&&generationCount>0)
		dropTable("Generation"+(generationCount-1));
	    statement.executeUpdate("UNLOCK TABLES");
	    statement.executeUpdate("FLUSH TABLES");
	}catch(SQLException e){
	    System.err.println("ConnectionString='"+connectionString+"'");
	    e.printStackTrace();
	    System.exit(-1);
	}
	long endTime = System.currentTimeMillis();
	timeUsed += endTime-startTime;
    }

    public void switchToNextGeneration(){

	if(!VP.STORETREE&&generationCount>0)
	    dropTable("Generation"+(generationCount-1));

	generationCount++;
	createTable();
    }
    
    public Iterator iterator(){
	return new VierGewinntDBIterator("Generation"+generationCount);
    }

    public void add(VierGewinntTurn t){
	//put the entry in the insert queue
	insertQ[insertQLoad++] = t;
	if(insertQLoad>=MAXQLOAD)
	    flushData();
    }

    private void flushData(){
	if(0==insertQLoad)
	    return;

	insertSQL.append("INSERT IGNORE INTO Generation");
	insertSQL.append(generationCount);
	insertSQL.append(" (data1,data2) VALUES ");
    
	for(int i=0;i<insertQLoad;i++){
	    VierGewinntTurn t = insertQ[i];
	    long[] data = t.getInternalData();
	    if(i>0)
		insertSQL.append(",(");
	     else insertSQL.append('(');
	    insertSQL.append(data[0]).append(',').append(data[1]);
	    insertSQL.append(')');
	}
	
	insertQLoad = 0;

	if(MULTITHREADED){
	    sqlExecuter.executeSQL(insertSQL.toString());
	}else{
	    long startTime = System.currentTimeMillis();

	    try{
		statement.executeUpdate(insertSQL.toString());
	    }catch(SQLException e){
		System.err.println("ConnectionString='"
				   +connectionString+"'");
		e.printStackTrace();
		System.exit(-1);
	    }

	    long endTime = System.currentTimeMillis();
	    timeUsed += endTime-startTime;
	}

	insertSQL.setLength(0);
    }
    
    public long size(){
	long ret;
	flushData();
	sqlExecuter.waitToFinish();
	
	long startTime = System.currentTimeMillis();
	try{
	    ResultSet result = statement.executeQuery(
			 "SELECT COUNT(*) FROM Generation"
			 +generationCount);
	    ret = result.getLong(1);
	}catch(SQLException e){
	    System.err.println("ConnectionString='"+connectionString+"'");
	    e.printStackTrace();
	    ret = 0;
	}
	long endTime = System.currentTimeMillis();
	timeUsed += endTime-startTime;

	return ret;
    }

    private class VierGewinntDBIterator implements Iterator{
	private String tableName = null;
	private ResultSet result = null;
	private boolean hasNext = false;
	private PreparedStatement query = null;
	private long offset = 0;

	public VierGewinntDBIterator(String tableName){
	    this.tableName = tableName;
	    offset = 0;

	    long startTime = System.currentTimeMillis();
	    
	    try{

		//lock until the last element has been read

		String sql = "SELECT data1, data2 FROM "+tableName
		    +" LIMIT ?,"+MAXFETCH;
		query = con.prepareStatement(sql);

	    }catch(SQLException e){
	    System.err.println("ConnectionString='"+connectionString+"'");
		e.printStackTrace();
		System.exit(-1);
	    }

	    getNextResults();

	    long endTime = System.currentTimeMillis();
	    timeUsed += endTime-startTime;
	}

	private void getNextResults(){
	    try{
		query.setLong(1,offset);
		offset += MAXFETCH;
		result = query.executeQuery();
		hasNext = result.next();
	    }catch(SQLException e){
	    System.err.println("ConnectionString='"+connectionString+"'");
		e.printStackTrace();
		System.exit(-1);
	    }
	}

	public boolean hasNext(){
	    return hasNext;
	}

	public Object next(){
	    try{
		long startTime = System.currentTimeMillis();

		long data1 = result.getLong(1);
		long data2 = result.getLong(2);
		if(!result.next())
		    getNextResults();
		long endTime = System.currentTimeMillis();
		timeUsed += endTime-startTime;

		return new VierGewinntTurn(data1,data2);
	    }catch(SQLException e){
	    System.err.println("ConnectionString='"+connectionString+"'");
		e.printStackTrace();
		System.exit(-1);
		return null;
	    }
	}

	public void remove(){
	    throw new RuntimeException("'remove' Not implemented");
	}

    }

/*
    private class VierGewinntDBIterator extends Thread implements Iterator{
	private String tableName = null;
	private ResultSet result = null;
	private ResultSet nextResult = null;
	private boolean hasNext = false;
	private PreparedStatement query = null;
	private long offset = 0;

	private boolean isFetching = false;

	public VierGewinntDBIterator(String tableName){
	    super();

	    this.tableName = tableName;
	    offset = 0;

	    try{
		String sql = "SELECT data1, data2 FROM "+tableName
		    +" LIMIT ?,"+MAXFETCH;
		query = con.prepareStatement(sql);
	    }catch(SQLException e){
		System.err.println("ConnectionString='"+connectionString+"'");
		e.printStackTrace();
		System.exit(-1);
	    }

	    setPriority(Thread.MIN_PRIORITY);
	    setDaemon(true); //application terminates if run hasn't terminated


	    prefetchResult();

	    getNextResult();
	    
	    if(MULTITHREADED)  start();
	}

	private void getNextResult(){
	    if(MULTITHREADED){
		System.out.println("starting(");

		synchronized(this){
		    if(isFetching){
			try{
			    wait();
			}catch(InterruptedException e){
			    e.printStackTrace();
			}
		    }
		}
		System.out.println("|");
	    

		try{
		    result = nextResult;
		    hasNext = result.next();
		}catch(SQLException e){
		    System.err.println("ConnectionString='"
				       +connectionString+"'");
		    e.printStackTrace();
		    System.exit(-1);
		}

		System.out.println("|");

		synchronized(this){
		    isFetching = true;
		    notifyAll();
		}

		synchronized(this){
		    try{
			wait();
		    }catch(InterruptedException e){
			e.printStackTrace();
		    }
		}
		System.out.println(")starting");
	    
	    }else{
		try{
		    result = nextResult;
		    hasNext = result.next();

		}catch(SQLException e){
		    System.err.println("ConnectionString='"
				       +connectionString+"'");
		    e.printStackTrace();
		    System.exit(-1);
		}

		if(hasNext)
		    prefetchResult();
	    }
	}

	public void run(){
	    while(hasNext){
		System.out.println("executing(");
		prefetchResult();
		System.out.println(")executing");
		    
		isFetching = false;
		synchronized(this){
		    notifyAll();
		}

		synchronized(this){
		    try{
			wait();
		    }catch(InterruptedException e){
			e.printStackTrace();
		    }
		}
	    }
	}

	private void prefetchResult(){
	    try{
		query.setLong(1,offset);
		offset += MAXFETCH;
		System.out.println("offset="+offset);
		nextResult = query.executeQuery();
		result = nextResult;
	    }catch(SQLException e){
		System.err.println("ConnectionString='"
				   +connectionString+"'");
		e.printStackTrace();
		System.exit(-1);
	    }
	}
	
	public boolean hasNext(){
	    return hasNext;
	}

	public Object next(){
	    try{
		long data1 = result.getLong(1);
		long data2 = result.getLong(2);
		if(!result.next())
		    getNextResult();
		System.out.println(data1);
		return new VierGewinntTurn(data1,data2);
	
	    }catch(SQLException e){
	    System.err.println("ConnectionString='"+connectionString+"'");
		e.printStackTrace();
		System.exit(-1);
		return null;
	    }
	}

	public void remove(){
	    throw new RuntimeException("'remove' Not implemented");
	}

    }
*/
}



